#include <SPAD/SYSLOG.H>

#include "IPGREG.H"
#include "IPG.H"

extern IO_STUB IPG_PACKET;

const HANDLE_OPERATIONS IPG_OPERATIONS = {
	SPL_X(SPL_IPG),
	KERNEL$NO_VSPACE_GET,
	KERNEL$NO_VSPACE_PUT,
	KERNEL$NO_VSPACE_MAP,
	KERNEL$NO_VSPACE_DMALOCK,
	KERNEL$NO_VSPACE_DMA64LOCK,
	KERNEL$NO_VSPACE_PHYSLOCK,
	KERNEL$NO_VSPACE_GET_PAGEIN_RQ,
	KERNEL$NO_VSPACE_GET_PAGE,
	KERNEL$NO_VSPACE_SWAP_OP,
	NULL,			/* clone */
	NULL,			/* lookup */
	NULL,			/* create */
	NULL, 			/* delete */
	NULL,			/* rename */
	NULL,			/* lookup_io */
	NULL,			/* instantiate */
	NULL,			/* leave */
	NULL,			/* detach */
	NULL,			/* close */
	KERNEL$NO_OPERATION,	/* read */
	KERNEL$NO_OPERATION,	/* write */
	KERNEL$NO_OPERATION,	/* aread */
	KERNEL$NO_OPERATION,	/* awrite */
	IPG_IOCTL,		/* ioctl */
	KERNEL$NO_OPERATION,	/* bio */
	IPG_PACKET,		/* pktio */
};

void IPG_ADD_RECV_PACKET(IPG *ipg, PACKET *p)
{
	VDMA64 vdma;
	VDESC desc;
	unsigned idx;
	IPG_RX_DESC *rx;
	p->addrlen = ETHERNET_ADDRLEN;
	desc.ptr = (unsigned long)p->data + LL_HEADER - sizeof(struct eth_header);
	desc.len = ipg->rx_pkt_size;
	desc.vspace = &KERNEL$VIRTUAL;
	RAISE_SPL(SPL_VSPACE);
	KERNEL$VIRTUAL.op->vspace_dma64lock(&desc, PF_WRITE, &vdma, (vspace_dma64unlock_t **)&p->sender_data);
	LOWER_SPL(SPL_IPG);
	idx = (ipg->first_recv + ipg->n_recv++) & ipg->recv_mask;
	ipg->recv_packets[idx] = p;
	rx = &RX_DESCS(ipg)[idx];
	rx->addr = __64CPU2LE(p->v.ptr = vdma.ptr);
	rx->status = 0;
}

static __finline__ void IPG_UNMAP_RECV_PACKET(IPG *ipg, PACKET *p)
{
	RAISE_SPL(SPL_VSPACE);
	((vspace_dma64unlock_t *)p->sender_data)(p->v.ptr);
	LOWER_SPL(SPL_IPG);
}

static __finline__ void IPG_UNMAP_TX_DESC(IPG *ipg, unsigned idx)
{
	RAISE_SPL(SPL_VSPACE);
	ipg->xmit_info[idx].unlock(__64LE2CPU(TX_DESCS(ipg)[idx].legacy.addr));
	LOWER_SPL(SPL_IPG);
}

static void IPG_SEND(IPG *ipg, PACKET *p)
{
	IPG_TX_DESC *tx;
	VDESC desc;
	VDMA64 vdma;
	unsigned pos;
	desc.ptr = (unsigned long)p->data + LL_HEADER - sizeof(struct eth_header);
	desc.len = p->data_length + sizeof(struct eth_header);
	desc.vspace = &KERNEL$VIRTUAL;
	RAISE_SPL(SPL_VSPACE);
	pos = (ipg->first_sent + ipg->n_sent) & ipg->sent_mask;
	KERNEL$VIRTUAL.op->vspace_dma64lock(&desc, PF_READ, &vdma, &ipg->xmit_info[pos].unlock);
	LOWER_SPL(SPL_IPG);
	set_tx:
	tx = &TX_DESCS(ipg)[pos];
	tx->legacy.addr = __64CPU2LE(vdma.ptr);
	tx->legacy.len = __16CPU2LE(vdma.len);
	tx->legacy.cmd = 0;
	ipg->xmit_info[pos].p = NULL;
	while (__unlikely(vdma.len > IPG_MAX_TX_DESC_SIZE)) {
		vspace_dma64unlock_t *unlock;
		tx->legacy.len = __16CPU2LE(IPG_MAX_TX_DESC_SIZE);
		vdma.len -= IPG_MAX_TX_DESC_SIZE;
		vdma.ptr += IPG_MAX_TX_DESC_SIZE;
		unlock = ipg->xmit_info[pos].unlock;
		ipg->xmit_info[pos].unlock = KERNEL$NULL_VSPACE_DMA64UNLOCK;
		pos = (pos + 1) & ipg->sent_mask;
		ipg->xmit_info[pos].unlock = unlock;
		goto set_tx;
	}
	if (__unlikely(p->v.len != 0)) {
		RAISE_SPL(SPL_VSPACE);
		pos = (pos + 1) & ipg->sent_mask;
		p->v.vspace->op->vspace_dma64lock(&p->v, PF_READ, &vdma, &ipg->xmit_info[pos].unlock);
		LOWER_SPL(SPL_IPG);
		if (__unlikely(!vdma.len)) {
			p->flags |= PKT_PAGING_ERROR;
			p->status = -EVSPACEFAULT;
			while (pos != ((ipg->first_sent + ipg->n_sent) & ipg->sent_mask)) {
				pos = (pos - 1) & ipg->sent_mask;
				IPG_UNMAP_TX_DESC(ipg, pos);
			}
			CALL_PKT(p);
			return;
		}
		p->v.ptr += vdma.len;
		p->v.len -= vdma.len;
		goto set_tx;
	}
	tx->legacy.cmd = IPG_TX_DESC_LEGACY_CMD_EOP | IPG_TX_DESC_LEGACY_CMD_IFCS | (IPG_TX_DESC_LEGACY_CMD_RS << (ipg->mac_type < MAC_82543));	/* !!! FIXME: don't need to report */
	ipg->xmit_info[pos].p = p;
	if (__likely(p->flags & PKT_OUTPUT_CHECKSUM)) {
		tx->legacy.cmd |= IPG_TX_DESC_LEGACY_CMD_IC;
		tx->legacy.css = CHECKSUM_FROM(p) + sizeof(struct eth_header);
		tx->legacy.cso = CHECKSUM_POS(p) + sizeof(struct eth_header);
	}
	ipg->n_sent = (pos + 1 - ipg->first_sent) & ipg->sent_mask;
	__barrier();
	IPG_WRITE(ipg, IPG_TDT + ipg->tdht_add, (ipg->first_sent + ipg->n_sent) & ipg->sent_mask);
}

DECL_IOCALL(IPG_PACKET, SPL_IPG, PACKET)
{
	RQ->status = -ENOSYS;
	RETURN_AST(RQ);
}

DECL_AST(IPG_PACKET_RETURNED, SPL_IPG, PACKET)
{
	IPG *ipg;
	PKT_AST_ENTER(RQ);
	ipg = RQ->sender_data;
	ipg->outstanding--;
	if (__unlikely(ipg->n_recv < ipg->recv_mask)) {
		RQ->flags &= PKT_REUSE_FLAGS;
		IPG_ADD_RECV_PACKET(ipg, RQ);
	} else {
		FREE_PACKET(RQ, NULL, SPL_IPG);
	}
	RETURN;
}

static void IPG_REAP_TX(IPG *ipg)
{
	unsigned pos = IPG_READ(ipg, IPG_TDH + ipg->tdht_add);
	if (__unlikely(pos > ipg->sent_mask)) {
		KERNEL$SYSLOG(__SYSLOG_HW_BUG, ipg->dev_name, "TX HEAD OUT OF RANGE: %X, MAX %X", pos, ipg->sent_mask);
	}
	while (ipg->first_sent != pos) {
		PACKET *p;
		if (__unlikely(!ipg->n_sent))
			KERNEL$SYSLOG(__SYSLOG_HW_BUG, ipg->dev_name, "TX HEAD OUT OF USED AREA: %X, HEAD %X", pos, ipg->first_sent);
		ipg->n_sent--;
		IPG_UNMAP_TX_DESC(ipg, ipg->first_sent);
		if (__likely((p = ipg->xmit_info[ipg->first_sent].p) != NULL)) {
			p->status = 0;
			CALL_PKT(p);
		}
		ipg->first_sent = (ipg->first_sent + 1) & ipg->sent_mask;
	}
}

static void IPG_PROCESS_RX(IPG *ipg)
{
	__barrier();
	while (__likely(ipg->n_recv)) {
		PACKET *p;
		IPG_RX_DESC *rx = &RX_DESCS(ipg)[ipg->first_recv];
		if (__unlikely(!(rx->status & IPG_RX_DESC_STATUS_DD)))
			break;
		p = ipg->recv_packets[ipg->first_recv];
		ipg->first_recv = (ipg->first_recv + 1) & ipg->recv_mask;
		ipg->n_recv--;
		IPG_UNMAP_RECV_PACKET(ipg, p);
		PKT_PREFETCH_1(p);
		p->flags |= PKT_INPUT_CHECKSUM;
		p->checksum.u = MKCHECKSUM(IPG_SUM_START, __16LE2CPU(rx->csum));
		p->fn = IPG_PACKET_RETURNED;
		p->sender_data = ipg;
		p->h = ipg->packet_input;
		p->id = IDTYPE_PCI | (ipg->id & IDTYPE_MASK);
		ipg->outstanding++;
		p->data_length = __16LE2CPU(rx->len) - sizeof(struct eth_header);
		if (__unlikely(p->data_length > ipg->mtu)) {
			p->status = -EIO;
			CALL_PKT(p);
			continue;
		}
		PKT_PREFETCH_2(p);
		CALL_IORQ(p, KERNEL$PKTIO);
	}
}

DECL_AST(IPG_IRQ, SPL_IPG, AST)
{
	IPG *ipg = GET_STRUCT(RQ, IPG, irq_ast);
	__u32 icr = IPG_READ(ipg, IPG_ICR);
	__debug_printf("irq: %08x\n", icr);
	if (__unlikely(ipg->mac_type == MAC_82547) || __unlikely(ipg->mac_type == MAC_82547_rev_2))
		IPG_WRITE(ipg, IPG_IMC, ~0);
	if (__likely(icr & IPG_ICR_TXDW))
		IPG_REAP_TX(ipg);
	if (__likely(icr & (IPG_ICR_RXDMT0 | IPG_ICR_RXTO)))
		IPG_PROCESS_RX(ipg);
	if (__unlikely(icr & (IPG_ICR_LSC | IPG_ICR_RXSEQ))) {
	}
	if (__unlikely(ipg->mac_type == MAC_82547) || __unlikely(ipg->mac_type == MAC_82547_rev_2))
		IPG_WRITE(ipg, IPG_IMS, INTERRUPT_MASK(ipg));
	ipg->irq_ctrl.eoi();
	RETURN;
}

void IPG_TIMER(TIMER *t)
{
	IPG *ipg = GET_STRUCT(t, IPG, timer);
	LOWER_SPL(SPL_IPG);
	SET_TIMER_NEVER(&ipg->timer);
}

void IPG_TEAR_DOWN_RECV_PACKETS(IPG *ipg)
{
	unsigned j;
	for (j = 0; j < ipg->n_recv; j++) {
		PACKET *p = ipg->recv_packets[(ipg->first_recv + j) & ipg->recv_mask];
		IPG_UNMAP_RECV_PACKET(ipg, p);
		FREE_PACKET(p, NULL, SPL_IPG);
	}
}

void IPG_TEAR_DOWN_SENT_PACKETS(IPG *ipg)
{
	unsigned i;
	for (i = 0; i < ipg->n_sent; ipg++) {
		PACKET *p;
		unsigned idx = (ipg->first_sent + i) & ipg->sent_mask;
		IPG_UNMAP_TX_DESC(ipg, idx);
		if (__likely((p = ipg->xmit_info[idx].p) != NULL)) {
			p->status = -EAGAIN;
			CALL_PKT(p);
		}
	}
}
