#ifndef __SPAD_SLAB_H
#define __SPAD_SLAB_H

#include <ARCH/SETUP.H>
#include <SYS/TYPES.H>
#include <ARCH/BSF.H>
#include <ARCH/SPL.H>
#include <STRING.H>

__BEGIN_DECLS

#if __DEBUG_MALLOC_SHORTAGE > 0
long int random(void);
#endif

extern __const__ int KERNEL$KERNEL;

struct __slhead;

#include <ARCH/SLAB.H>

#ifndef __static__
#define __static__ static
#endif

/* !!! WARNING: constructor cannot contruct first pointer of an object --- it is   used as a free-list link */

int KERNEL$SLAB_ENTRIES_PER_PAGE(unsigned size, unsigned align);
void KERNEL$SLAB_INIT(struct __slhead *__h, unsigned __size, unsigned __align, int __alloc_type, void (*__ctor)(struct __slhead *__g, void *__o), long *__limit, char *__name);
void KERNEL$SLAB_DESTROY(struct __slhead *__h);
int KERNEL$SLAB_EMPTY(struct __slhead *__h);
int KERNEL$SLAB_RESERVE(struct __slhead *__h, unsigned __n);
int KERNEL$SLAB_REAP(void);
void KERNEL$SLABS_DUMP(void);

struct __slentry *__MALLOC_ATTR__ KERNEL$__SLAB_ALLOC(struct __slhead *__h);
void KERNEL$__SLAB_FREE_PAGE(struct __slpage *__p);
void KERNEL$__SLAB_PARTIAL_PAGE(struct __slpage *__p);

void *__MALLOC_ATTR__ KERNEL$__CALLOC(size_t __s);
void KERNEL$__MALLOC_ERROR(void);

#if defined(__NO_INLINES)
void *__MALLOC_ATTR__ calloc(size_t __n, size_t __size);
#else
__static__ __finline__ void *__MALLOC_ATTR__ calloc(size_t __n, size_t __size)
{
	size_t __ss;
#if __DEBUG_MALLOC_SHORTAGE > 0
	if (__unlikely(KERNEL$KERNEL) && __unlikely((random() & 255) < __DEBUG_MALLOC_SHORTAGE)) return NULL;
#endif
	if (__is_constant(__size)) {
		if (__size > 1 && __unlikely(__n > (size_t)-1 / __size)) {
			goto __me;
		}
		__ss = __n * __size;
	} else if (__is_constant(__n)) {
		if (__n > 1 && __unlikely(__size > (size_t)-1 / __n)) {
			goto __me;
		}
		__ss = __n * __size;
	} else {
		__calloc_multiply
	}
	return KERNEL$__CALLOC(__ss);
	__me:
	KERNEL$__MALLOC_ERROR();
	return NULL;
}
#endif

#if defined(__NO_INLINES) || defined(__COMMON_C)
void *__MALLOC_ATTR__ malloc(size_t __size);
#else
void *__MALLOC_ATTR__ ___malloc_alias(size_t __size);
void *__MALLOC_ATTR__ ___malloc_const(size_t __size_m1, unsigned __log2);
__static__ __finline__ void *__MALLOC_ATTR__ malloc(size_t __size)
{
	if (__is_constant(__size)) {
		size_t __x_size = (__size - 1) | (__MIN_SLAB_SIZE - 1);
		return ___malloc_const(__x_size, __BSR_CONST(__x_size));
	} else {
		return ___malloc_alias(__size);
	}
}
#endif

void *__MALLOC_ATTR__ realloc(void *__s, size_t __size);
void *__MALLOC_ATTR__ reallocf(void *__s, size_t __size);
size_t __alloc_size(void *__s);
void *__MALLOC_ATTR__ memalign(size_t __align, size_t __s);
void *__MALLOC_ATTR__ valloc(size_t __size);
void *__MALLOC_ATTR__ __sync_malloc(size_t __size);
void *__MALLOC_ATTR__ __sync_calloc(size_t __n, size_t __size);
void *__MALLOC_ATTR__ __sync_realloc(void *__s, size_t __size);
void *__MALLOC_ATTR__ __sync_reallocf(void *__s, size_t __size);
void *__MALLOC_ATTR__ __sync_memalign(size_t __align, size_t __size);

void free(void *__s);

__END_DECLS

#endif
